"""Top-level package for Another Sudoku Library."""

__author__ = """jc400"""
__email__ = " "
__version__ = "0.1.0"

from .utils import getEmptyBoard, copyBoard, getBitmap, fullGen, shuffle
from .solve import checkComplete, checkConsistent, check, solve
from .generate import getPuzzle



import pytest
from another_sudoku_library.solve import _getRowVals, _getColVals, _getSqrVals, solve, check
from . import sample_boards


def test_getRowVals():
    assert _getRowVals(sample_boards.t1, 3, 7, inclusive=False) == [6, 5, 7]


def test_getColVals():
    assert _getColVals(sample_boards.t1, 1, 1, inclusive=False) == [2, 4, 5]


def test_getSqrVals():
    assert _getSqrVals(sample_boards.t1, 4, 4, inclusive=False) == [5, 7]


@pytest.mark.parametrize(
    "pz",
    (
        {"problem": sample_boards.t2_problem, "solution": sample_boards.t2_solution},
        {"problem": sample_boards.t3_problem, "solution": sample_boards.t3_solution},
    ),
)
def test_solve(pz):
    assert solve(pz["problem"]) == pz["solution"]


@pytest.mark.parametrize("solution", (
    sample_boards.t2_solution,
    sample_boards.t3_solution,
)
)
def test_check(solution):
    assert check(solution) == True

    solution[0][0] = (solution[0][0] % 9) + 1   # make a cell incorrect
    assert check(solution) == False

    solution[0][0] = 0                          # make a cell empty
    assert check(solution) == False



"""Function to generate a new Sudoku puzzle"""

import random
from .utils import fullGen, exclusiveGen, getEmptyBoard, copyBoard
from .solve import _generateCache, _uniqueCheck, checkComplete, checkConsistent, solve


def _picker(board, y, x, cache=None):
    """Returns value for cell on board. First checks for existing value, 
    
    then checks if there's only one possible value, finally chooses randomly
    among valid possibilities.
    """
    if board[y][x] != 0:
        return board[y][x]

    if cache == None:
        cache = _generateCache(board)

    solve = _uniqueCheck(board, y, x, cache=cache)
    if solve != 0:
        return solve

    else:
        options = list(cache[y][x])
        return options[random.randint(0, len(options) - 1)]


def _generate():
    """Randomly generates a completely solved, logically consistent Sudoku board. 
    
    This doesn't work every time, so 'while not done:' loop continues trying until a 
    consistent board is generated.
    """
    done = False
    while not done:
        n = getEmptyBoard()
        for y, x in fullGen():
            n[y][x] = _picker(n, y, x)

            if y > 5:  # heuristic: try to solve once board is mostly full
                n = solve(n, nest=4)
                if checkComplete(n):
                    break

        if checkConsistent(n):
            done = True
    return n


def _removable(board, y, x, nest=4):
    """Tests if board can be solved without given cell. Returns bool. """
    test = copyBoard(board)
    test[y][x] = 0
    test = solve(test, nest=nest)  
    return checkComplete(test) and checkConsistent(test)


def _rm(wb, forbid, ry, rx, nest=4):
    """Checks if cell is removable or not. If yes, removes it from wb. If not, adds 
    
    that cell to the forbid list. Return True if removed, False if not. Nest arg controls
    how hard it tries to solve puzzle without the given cell--higher nest is slower, but
    removes more cells.
    """
    if wb[ry][rx] == 0 or forbid[ry][rx] == 1: # skip
        return 0

    if _removable(wb, ry, rx, nest=nest): 
        wb[ry][rx] = 0
        return True
    else:
        forbid[ry][rx] = 1
        return False


def _carve(board, count=60):
    """Takes a full board, and removes cells so that puzzle is still solveable. 
    
    Removes up to count cells. Note that there is a limit, you typically can't remove
    more than ~60 cells without making puzzle unsolveable. More cells removed = harder
    puzzle.
    """
    wb = copyBoard(board)
    forbid = getEmptyBoard()  # cache cells that are necessary to solve puzzle
    removes = 0
    attempts = 0

    # stage one: try removing random coords
    while removes < count and attempts < 50:
        attempts += 1
        ry = random.randint(0, 8)
        rx = random.randint(0, 8)
        if _rm(wb, forbid, ry, rx): 
            removes += 1

    # stage two: crawl entire board and try remove
    if removes < count:
        for ry, rx in exclusiveGen():
            if _rm(wb, forbid, ry, rx):
                removes += 1
                if removes == count:
                    break

    return wb


def getPuzzle(diff=40):
    """Generates completed puzzle, then carves off cells to create puzzle. 
    
    Diff argumnet is the number of cells to remove. 60 is hard, 25 is easy"""
    return _carve(_generate(), count=diff)



"""Functions to solve a sudoku puzzle, and/or check correctness"""

from .utils import rowGen, colGen, sqrGen, fullGen, getEmptyBoard, copyBoard


def _getRowVals(board, y, x, inclusive=True):
    """Returns list with the values of already set cells in row"""
    out = []
    for y1, x1 in rowGen(y, x, inclusive=inclusive):
        if board[y1][x1] == 0:
            continue
        elif type(board[y][x]) == int:
            out.append(board[y1][x1])
        else:
            out += board[y1][x1]
    return out


def _getColVals(board, y, x, inclusive=True):
    """Returns list with the values of already set cells in col"""
    out = []
    for y1, x1 in colGen(y, x, inclusive=inclusive):
        if board[y1][x1] == 0:
            continue
        elif type(board[y][x]) == int:
            out.append(board[y1][x1])
        else:
            out += board[y1][x1]
    return out


def _getSqrVals(board, y, x, inclusive=True):
    """Returns list with the values of already set cells in sqr"""
    out = []
    for y1, x1 in sqrGen(y, x, inclusive=inclusive):
        if board[y1][x1] == 0:
            continue
        elif type(board[y][x]) == int:
            out.append(board[y1][x1])
        else:
            out += board[y1][x1]
    return out


def _getPoss(board, y, x):
    """Returns set of possible values for a cell. Returns single-value set if cell is decided."""
    if board[y][x] != 0:
        return [board[y][x]]

    return (
        {1, 2, 3, 4, 5, 6, 7, 8, 9}
        - set(_getRowVals(board, y, x))
        - set(_getColVals(board, y, x))
        - set(_getSqrVals(board, y, x))
    )


def _countZeros(board):
    """Returns count of zeros (eg carved cells) in a provided board.

    Should work on board serialized with stringify(), and also full list board"""

    if type(board) == str:
        return board.count("0")

    zeros = 0
    for y, x in fullGen():
        if board[y][x] == 0:
            zeros += 1
    return zeros


def _generateCache(board):
    """Returns parallel cache board. Each cell in cache board holds the set of 
    
    possible values for that cell.
    """
    out = getEmptyBoard()
    for y, x in fullGen():
        out[y][x] = _getPoss(board, y, x)
    return out


def _uniqueCheck(board, y, x, cache=None):
    """Tries to solve cell. Return 1-9 if found, 0 if inconclusive, -1 if no possibilities.

    Looks like this is 16x faster w cache provided. Also if we calc cache here, we're doing a lot of
    work to get possibilities for cells we won't access.

    Works by getting the set of possible values for the cell, then comparing with the 
    possible values for every other cell in its row/col/sqr.
    """

    # run cheap, naive checks first
    cellPoss = _getPoss(board, y, x)
    if len(cellPoss) == 1:  # if cell has one poss, must be answer
        return cellPoss.pop()
    elif len(cellPoss) == 0:  # if no poss for cell, must be error in board
        return -1

    # if no cache, we do have to generate one
    if cache == None:
        cache = _generateCache(board)

    # compare cell possibilities to unit possibilities. Looking for UNIQUE poss in cell
    if cellPoss - set(_getRowVals(cache, y, x, inclusive=False)):
        return (cellPoss - set(_getRowVals(cache, y, x, inclusive=False))).pop()
    if cellPoss - set(_getColVals(cache, y, x, inclusive=False)):
        return (cellPoss - set(_getColVals(cache, y, x, inclusive=False))).pop()
    if cellPoss - set(_getSqrVals(cache, y, x, inclusive=False)):
        return (cellPoss - set(_getSqrVals(cache, y, x, inclusive=False))).pop()

    # if inconclusive, just return nothing
    return 0


def solve(board, nest=0):
    """This is best-effort solve. Tries its best, but may return an incomplete or inconsistent

    board. If error, should have that cell as -1.
    """
    wb = copyBoard(board)  # pure function--dont change original board arg
    c = _generateCache(wb)
    solved = False
    changed = True

    while not solved and changed:
        changed = False
        for y, x in fullGen():
            if wb[y][x] == 0:  # only check unsolved cells
                wb[y][x] = _uniqueCheck(wb, y, x, cache=c)
                if wb[y][x] > 0:
                    c = _generateCache(wb)  # regen cache if we hit an answer
                    changed = True

        if checkComplete(wb):
            solved = True

    # if above checks don't solve puzzle, try to brute force
    if not solved and nest < 3:
        lpy, lpx = 0, 0  # get lynchpin, cell with most possibilities.
        for y, x in fullGen():
            if len(c[y][x]) > len(c[lpy][lpx]):
                lpy, lpx = y, x

        for value in c[lpy][lpx]:   # check each possible value for lynchpin cell
            testWB = copyBoard(wb)
            testWB[lpy][lpx] = value
            testWB = solve(testWB, nest=nest + 1)  # control recursion with nest arg

            if checkComplete(testWB) and checkConsistent(testWB):
                if not solved:
                    wb = testWB
                    solved = True
                else:  # board already solved means multiple valid solutions. Fail.
                    wb[0][0] = 0

    return wb


def checkComplete(board):
    """Naive check whether each cell has been filled (or if its still 0)"""
    for y, x in fullGen():
        if board[y][x] == 0:
            return False
    return True


def checkConsistent(board):
    """True if consistent, else false. Works by checking whether a cell 
    
    matches another cell in its unit (row/column/square)
    """
    if not checkConsistentCheap(board):
        return False

    for y, x in fullGen():
        if board[y][x] == 0:
            continue

        temp = (
            _getRowVals(board, y, x, inclusive=False)
            + _getColVals(board, y, x, inclusive=False)
            + _getSqrVals(board, y, x, inclusive=False)
        )
        if board[y][x] in temp:
            return False

    return True


def checkConsistentCheap(board):
    """Return true if board is logically consistent, false otherwise.
    
    This is heuristic, simply checks if cell has -1 since this indicates 
    and error
    """
    for y, x in fullGen():
        if board[y][x] == -1:
            return False
    return True


def check(board):
    """Checks that board is both complete and consistent. Returns False if not."""
    return checkComplete(board) and checkConsistent(board)



from another_sudoku_library.generate import getPuzzle
from another_sudoku_library.utils import fullGen
from another_sudoku_library.solve import checkConsistent


def test_getPuzzle():
    result = getPuzzle()
    assert checkConsistent(result) == True
    for y, x in fullGen():
        assert result[y][x] in range(0, 10)


"""Utilities for accessing and manipulating sudoku boards """

import random


def getEmptyBoard():
    """Returns empty board, a 9x9 2D list structure. Empty cells are represented as zeros."""
    out = [
        [0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0, 0],
    ]
    return out


def printBoard(board, hidezeros=True):
    """Pretty print a board to console"""
    for y, x in fullGen():
        # print every number
        if board[y][x] == 0 and hidezeros:
            print(" ", end=" ")
        else:
            print(board[y][x], end=" ")

        # formatting & separators
        if x == 8:
            print()
            if y == 2 or y == 5:
                print("----------------------")

        if x == 2 or x == 5:
            print("| ", end="")


def copyBoard(board):
    """Deep copy--creates and returns copy as an entirely new board structure."""
    copy = getEmptyBoard()
    for y, x in fullGen():
        copy[y][x] = board[y][x]
    return copy


def getBitmap(board):
    """Returns board with 1 for all provided cells, 0 for undecided cells.

    This is useful for implementing a UI, for instance to separately style provided cells,
    or to prevent user from changing provided cells.
    """
    out = getEmptyBoard()
    for y, x in fullGen():
        if board[y][x] != 0:
            out[y][x] = 1
    return out


def getOrigPermutation(board, bitmap):
    """Returns original problem board based on bitmap"""
    out = getEmptyBoard()
    for y, x in fullGen():
        out[y][x] = board[y][x] * bitmap[y][x]
    return out


def rotateNormal(board):
    """Rotates board so heaviest side is on bottom. Goal is consistency--identical puzzles should end up
    the same, regardless of initial orientation (identical meaning *normal* versions of the puzzle are identical).

    I want to take a given puzzle and rotate it, so that different permutations always end up in the same
    orientation. I cannot use the values of the cells, because these will also be normalized. So, I have to
    consider only the *locations* of filled vs unfilled cells--we're essentially figuring out which side
    is the heaviest.

    If I only counted all the values on a given side (eg the top is every cell in first three rows), sides
    would frequently have the same weight. So, we scale each row (and each square) slightly differently.
    The scaling should be the same between different sides, so the same side gets the same weighted score
    regardless of whether it's currently on top, left, bottom or right. But specific *cells* within that side
    are weighted differently, to try to give differentiation.

    We still have problem of a completely symmetrical puzzle being non-rotateable. But giving these funky
    weights decreases the chance of that (hopefully)

    """
    a, b, c, d = 0, 0, 0, 0  # these hold sum of weights for each side

    t1 = sum(1.114 for y, x in rowGen(0, 4) if board[y][x] != 0)
    t2 = sum(1.326 for y, x in rowGen(1, 4) if board[y][x] != 0)
    t3 = sum(1.453 for y, x in rowGen(2, 4) if board[y][x] != 0)
    t4 = sum(1.242 for y, x in sqrGen(0, 0) if board[y][x] != 0)
    t5 = sum(1.141 for y, x in sqrGen(0, 3) if board[y][x] != 0)
    t6 = sum(1.242 for y, x in sqrGen(0, 6) if board[y][x] != 0)
    a = t1 + t2 + t3 + t4 + t5 + t6

    t1 = sum(1.114 for y, x in colGen(4, 8) if board[y][x] != 0)
    t2 = sum(1.326 for y, x in colGen(4, 7) if board[y][x] != 0)
    t3 = sum(1.453 for y, x in colGen(4, 6) if board[y][x] != 0)
    t4 = sum(1.242 for y, x in sqrGen(0, 8) if board[y][x] != 0)
    t5 = sum(1.141 for y, x in sqrGen(3, 8) if board[y][x] != 0)
    t6 = sum(1.242 for y, x in sqrGen(6, 8) if board[y][x] != 0)
    b = t1 + t2 + t3 + t4 + t5 + t6

    t1 = sum(1.114 for y, x in rowGen(8, 4) if board[y][x] != 0)
    t2 = sum(1.326 for y, x in rowGen(7, 4) if board[y][x] != 0)
    t3 = sum(1.453 for y, x in rowGen(6, 4) if board[y][x] != 0)
    t4 = sum(1.242 for y, x in sqrGen(8, 0) if board[y][x] != 0)
    t5 = sum(1.141 for y, x in sqrGen(8, 3) if board[y][x] != 0)
    t6 = sum(1.242 for y, x in sqrGen(8, 6) if board[y][x] != 0)
    c = t1 + t2 + t3 + t4 + t5 + t6

    t1 = sum(1.114 for y, x in colGen(4, 0) if board[y][x] != 0)
    t2 = sum(1.326 for y, x in colGen(4, 1) if board[y][x] != 0)
    t3 = sum(1.453 for y, x in colGen(4, 2) if board[y][x] != 0)
    t4 = sum(1.242 for y, x in sqrGen(0, 0) if board[y][x] != 0)
    t5 = sum(1.141 for y, x in sqrGen(3, 0) if board[y][x] != 0)
    t6 = sum(1.242 for y, x in sqrGen(6, 0) if board[y][x] != 0)
    d = t1 + t2 + t3 + t4 + t5 + t6

    if a > b and a > c and a > d:
        return board
    elif b > a and b > c and b > d:
        return rotate(board)
    elif c > a and c > b and c > d:
        return rotate(board, rotates=2)
    elif d > a and d > b and d > c:
        return rotate(board, rotates=3)
    else:
        print("error, two values collide")
        print("a: ", a, "b: ", b, "c: ", c, "d: ", d)
        return getEmptyBoard()


def orderNormal(board):
    """Re-orders values in puzzle, without changing logic. Goal is consistency"""

    temp = copyBoard(board)
    for digit in range(1, 9):  # go through 1 - 9, to ensure puzzle is in order
        for y, x in fullGen():  # check through entire board
            if (
                temp[y][x] < digit
            ):  # if cell is 0 or smaller number, ignore since already in order
                continue
            elif (
                temp[y][x] == digit
            ):  # if next cell we come to is our digit, break since already in order
                break
            else:  # if cell is out of order, swap all occurences of number w digit
                old = 1 * temp[y][x]  # get old, out of order num
                for y1, x1 in fullGen():  # new loop through board
                    if temp[y1][x1] == old:  # swap
                        temp[y1][x1] = digit
                    elif temp[y1][x1] == digit:
                        temp[y1][x1] = old
                break
    return temp


def normalize(board):
    """returns normalized form of board--all permutations should result in same normal form"""
    return orderNormal(rotateNormal(board))


def shuffle(board):
    """Shuffle a normalized board to return a random permutation"""
    digits = [1, 2, 3, 4, 5, 6, 7, 8, 9]
    out = copyBoard(board)

    for old in range(1, 10):
        digit = digits[random.randint(0, len(digits) - 1)]
        digits.remove(digit)
        for y, x in fullGen():
            if out[y][x] == old:
                out[y][x] = digit
            elif out[y][x] == digit:
                out[y][x] = old

    r = random.randint(0, 3)
    return rotate(out, rotates=r)


def stringify(board):
    """Serializes board (a 2D list) into a flat string separated by semicolons"""
    outString = ""
    for y, x in fullGen():
        outString += str(board[y][x])
        if x == 8 and y != 8:
            outString += ";"
    return outString


def unstringify(boardString):
    """Unserializes board (a 2D list) from a flat string"""
    out = getEmptyBoard()
    temp = boardString.split(";")
    for y, x in fullGen():
        out[y][x] = int(temp[y][x])
    return out


def rowGen(y, x, inclusive=True):
    """Return generator that gives coord tuples (y,x) for the row of the argument coordinate.

    By default, generator returns the argument coordinate. Set inclusive=False to exclude
    the argument.
    """
    return ((y, x1) for x1 in range(9) if (inclusive) or (x1 != x))


def colGen(y, x, inclusive=True):
    """Return generator that gives coord tuples (y,x) for the column of the argument coordinate.

    By default, generator returns the argument coordinate. Set inclusive=False to exclude
    the argument.
    """
    return ((y1, x) for y1 in range(9) if (inclusive) or (y1 != y))


def sqrGen(y, x, inclusive=True):
    """Return generator that gives coord tuples (y,x) for the square of the argument coordinate.

    By default, generator returns the argument coordinate. Set inclusive=False to exclude
    the argument.
    """
    cY = (y // 3) * 3  # calculate top-left corner of square, for
    cX = (x // 3) * 3  # the provided cell
    return (
        (y1, x1)
        for y1 in range(cY, cY + 3)
        for x1 in range(cX, cX + 3)
        if (inclusive) or (y1 != y) or (x1 != x)
    )


def fullGen():
    """Returns generator that gives coord tuples (y,x) for whole board"""
    return ((y1, x1) for y1 in range(9) for x1 in range(9))


def exclusiveGen():
    """Returns generator that gives coord tuples (y,x) for whole board. Coords are

    given in order to get maximum distance between consecutive tuples."""
    a = lambda x, y: ((x * 3 + x // 3 + y) % 9)
    return ((a(y1, x1), x1) for y1 in range(9) for x1 in range(9))


def funnyGen():
    """Returns generator that gives coord tuples (y,x) for whole board. Coords fill

    out one square at a time."""
    xs = [0, 1, 2, 0, 1, 2, 0, 1, 2]
    ys = [0, 1, 2, 2, 0, 1, 1, 2, 0]
    for sqr in range(9):
        cornerY = ys[sqr] * 3
        cornerX = xs[sqr] * 3
        for cel in range(9):
            yield cornerY + ys[cel], cornerX + xs[cel]


def rotate(start, rotates=1):
    """spins board clockwise 90 degrees for each 'rotate'. Returns new board"""
    workboard = copyBoard(start)
    for i in range(rotates):
        temp = getEmptyBoard()
        for y, x in fullGen():
            y1 = x
            x1 = abs(8 - y)
            temp[y1][x1] = workboard[y][x]
        workboard = temp
    return workboard
